import logging as lg
import os
from curses import COLOR_MAGENTA
from multiprocessing import Pool
from typing import Any, Callable, Hashable, Iterable, Literal

import polars as pl
from rdkit.Chem import Mol

from fairfetched.utils.polars import (
    map_batches_unique,
    map_elements_cached,
    map_elements_pooled_cached,
    map_elements_unique,
    map_elements_unique_pooled,
)

from .compounds import (
    _safe_inchi_to_mol,
    _safe_smiles_to_mol,
    fuse_pipeline,
    inchi_to_inchikey,
    mol_to_strings,
    mols_to_inchikeys,
    mols_to_inchis,
    mols_to_inchis_and_auxinfo,
    mols_to_kekulised_smiles,
    standardise_aeffect,
    standardise_papyrus,
)

DFLike = pl.DataFrame | pl.LazyFrame


def with_standardised_mols(
    df,
    *standardisation_functions,
    start_col: str = "smiles",
    str_to_mol_function: Callable[[str], Mol | None] | None = None,
    mol_to_repr_function: Callable[[Mol | None], str | tuple[str, ...] | Any]
    | None = None,
    new_col_name: str = "standardised",
    final_dtype: pl.DataType | pl.DataTypeExpr | None = pl.Object,  # ty: ignore[invalid-parameter-default]
):
    """Example usage:
    ```
        from fairfetched.standardisation import compounds as ms
        from fairfetched.standardisation.pipe import smiles_inchi_auxinfo_inchikey_to_three_cols
        df.pipe(
            with_standardised_mols,
            ms.standardise_papyrus,
            start_col="inchi",
            new_col_name="smiles_inchi_auxinfo_inchikey",
            final_dtype=pl.Array(inner=pl.Utf8, shape=3),
            str_to_mol_function=ms._safe_inchi_to_mol,
            mol_to_repr_function=ms.mol_to_strings,
        ).pipe(smiles_inchi_auxinfo_inchikey_to_three_cols, )

        #is roughly equivalent to
        df.with_columns(
            smiles_inchi_auxinfo_inchikey=pl.col("inchi")
            .map_elements(ms._safe_inchi_to_mol, return_dtype=pl.Object)
            .map_elements(ms.standardise_papyrus, return_dtype=pl.Object)
            .map_elements(ms.mol_to_strings, return_dtype=pl.Array(inner=pl.Utf8, shape=3))
        )
        #but should be faster as it reduces multiple function call overhead
        # and evaluates only for unique molecules.
    ```
    """
    if str_to_mol_function:
        standardisation_functions = (str_to_mol_function, *standardisation_functions)
    if mol_to_repr_function:
        standardisation_functions = (*standardisation_functions, mol_to_repr_function)
    return df.pipe(
        map_elements_unique,
        fuse_pipeline(*standardisation_functions),  # ty: ignore[invalid-argument-type]
        col=start_col,
        alias=new_col_name,
        return_dtype=final_dtype,
    )


def map_pipeline(
    df: pl.LazyFrame,
    *functions: Callable[[Mol | Any], Mol | Any],
    col="smiles",
    alias="standardised_form",
    return_dtype: pl.DataType | pl.DataTypeExpr | None = pl.Object,  # ty: ignore[invalid-parameter-default]
    pooled=True,
) -> pl.LazyFrame:
    """element-wise map a pipeline of functions to unique elements. uses a join at the end to combine into the lazyframe"""
    if not isinstance(df.select(col).collect_schema().dtypes()[0], Hashable):
        return df.with_columns(
            pl.col(col)
            .map_elements(
                fuse_pipeline(*functions),
                return_dtype=return_dtype,
            )
            .alias(alias)
        )
    map_fx = map_elements_cached
    if pooled:
        map_fx = map_elements_pooled_cached
    return df.pipe(
        map_fx,
        fuse_pipeline(*functions),
        col=col,
        alias=alias,
        return_dtype=return_dtype,
    )


def map_pipeline_unique_join(
    df: pl.LazyFrame,
    *functions: Callable[[Mol | Any], Mol | Any],
    col="smiles",
    alias="standardised_form",
    return_dtype: pl.DataType | pl.DataTypeExpr | None = pl.Object,  # ty: ignore[invalid-parameter-default]
) -> pl.LazyFrame:
    """element-wise map a pipeline of functions to unique elements. uses a join at the end to combine into the lazyframe"""

    return df.pipe(
        map_elements_unique,
        fuse_pipeline(*functions),
        col=col,
        alias=alias,
        return_dtype=return_dtype,
    )


def map_pipeline_unique_join_pooled(
    df: pl.LazyFrame,
    *functions: Callable[[Mol | Any], Mol | Any],
    col="smiles",
    alias="standardised_form",
    return_dtype: pl.DataType | pl.DataTypeExpr | None = pl.Object,  # ty: ignore[invalid-parameter-default]
) -> pl.LazyFrame:
    """element-wise map a pipeline of functions to unique elements. uses a join at the end to combine into the lazyframe"""
    return df.pipe(
        map_elements_unique_pooled,
        fuse_pipeline(*functions),
        col=col,
        alias=alias,
        return_dtype=return_dtype,
    )


def polars_pipe_pipeline(df, *pipe_functions):
    for fnx in pipe_functions:
        df = df.pipe(fnx)
    return df


def unnest_molstrings(
    df: pl.LazyFrame,
    col: str = "smiles_inchi_auxinfo_inchikey",
    rename_old_suffix: str = "_original",
) -> pl.LazyFrame:
    return (
        df.rename(
            {
                c: f"{c}{rename_old_suffix}"
                for c in ["smiles", "inchi", "inchi_auxinfo", "inchikey"]
                if c in df.collect_schema().keys()
            },
            strict=False,
        )
        .with_columns(
            pl.col(col)
            .arr.to_struct(fields=["smiles", "inchi", "inchi_auxinfo", "inchikey"])
            .struct.unnest()
        )
        .select(
            pl.col(df.collect_schema().names()),
            pl.exclude(df.collect_schema().names()),
        )
        .drop(col)
    )


def with_mols_from_inchi_pool(df: DFLike, alias="mol", parallel=True) -> DFLike:
    """
    for ~1M mols: 1h07m for fully linear one, 15m for this impl.
    """
    if parallel:
        unique_inchis = df.lazy().select(pl.col("inchi")).unique().collect()["inchi"]  # ty: ignore
        with Pool() as p:
            res = p.map(_safe_inchi_to_mol, unique_inchis)
        mols_df = pl.DataFrame(
            [
                pl.Series(unique_inchis, dtype=pl.Utf8).alias("inchi"),
                pl.Series(res, dtype=pl.Object).alias(alias),
            ]
        )
        return df.lazy().join(
            mols_df.lazy(),
            on="inchi",
            how="left",
            maintain_order="left",
        )
    return df.lazy().join(
        df.lazy()
        .select(pl.col("inchi"))
        .unique()
        .with_columns(
            pl.col("inchi")
            .map_elements(_safe_inchi_to_mol, return_dtype=pl.Object)
            .alias(alias)
        ),
        on="inchi",
        how="left",
        maintain_order="left",
    )


def with_mols_from_inchi(df: pl.DataFrame, alias="mol", parallel=True) -> DFLike:
    """
    Parallel InChI-to-mol conversion using ThreadPoolExecutor (safe with JAX).
    Returns a new DataFrame with a column of RDKit Mol objects from InChI strings.
    """
    if parallel:
        from concurrent.futures import ThreadPoolExecutor

        unique_inchis = (
            df.select(pl.col("inchi"))
            .unique()
            .get_column("inchi")  # ignore ty[not-subscriptable]
        )
        with ThreadPoolExecutor(
            max_workers=round((os.cpu_count() or 1) * 0.9)
        ) as executor:
            res = list(executor.map(_safe_inchi_to_mol, unique_inchis))
        mols_df = pl.DataFrame(
            [
                pl.Series(unique_inchis, dtype=pl.Utf8).alias("inchi"),
                pl.Series(res, dtype=pl.Object).alias(alias),
            ]
        )
        return df.lazy().join(
            mols_df.lazy(),
            on="inchi",
            how="left",
            maintain_order="left",
        )
    return df.lazy().join(
        df.lazy()
        .select(pl.col("inchi"))
        .unique()
        .with_columns(
            pl.col("inchi")
            .map_elements(_safe_inchi_to_mol, return_dtype=pl.Object)
            .alias(alias)
        ),
        on="inchi",
        how="left",
        maintain_order="left",
    )


def with_clean_mols_from_smiles(df: DFLike, alias="mol") -> DFLike:
    return df.with_columns(
        pl.col("smiles")
        .map_elements(_safe_smiles_to_mol, return_dtype=pl.Object, strategy="threading")
        .alias("mol")
    )


def _clean_mols_og_smiles_df(smiles: Iterable[str]) -> pl.DataFrame:
    with Pool() as p:
        res = p.map(_safe_smiles_to_mol, smiles)
    return pl.DataFrame(
        [
            pl.Series(smiles, dtype=pl.Utf8).alias("smiles"),
            pl.Series(res, dtype=pl.Object).alias("mol"),
        ]
    )


def _series_mols_to_string_representation(
    mols: pl.Series,
    type: Literal["smiles", "inchi", "inchi_and_auxinfo", "inchikey"] = "smiles",
) -> pl.Series:
    match type:
        case "smiles":
            func = mols_to_kekulised_smiles
            dtype = pl.Utf8
        case "inchi":
            func = mols_to_inchis
            dtype = pl.Utf8
        case "inchi_and_auxinfo":
            func = mols_to_inchis_and_auxinfo
            dtype = pl.Struct({"inchi": pl.Utf8, "inchi_auxinfo": pl.Utf8})
        case "inchikey":
            func = mols_to_inchikeys
            dtype = pl.Utf8
        case _:
            raise NotImplementedError
    return pl.Series(name=type, values=func(mols), dtype=dtype)


def _unique_with_clean_string_representations(df: DFLike) -> DFLike:
    """cleans and standardises molecules, replacing original string
    representations (which are tagged _papyrus) and adding new ones
    without the suffix. This is to ensure that all string representations
    are consistent and can be used for downstream tasks without confusion."""
    df = df.pipe(with_clean_mols_from_smiles).rename(
        {k: f"{k}_papyrus" for k in ["smiles", "inchi", "inchi_auxinfo", "inchikey"]},
        strict=False,
    )
    mols = df.lazy().select("mol").collect().get_column("mol")
    return df.with_columns(
        _series_mols_to_string_representation(mols, type="smiles"),
        _series_mols_to_string_representation(
            mols, type="inchi_and_auxinfo"
        ).struct.unnest(),
    ).with_columns(
        pl.col("inchi")
        .map_elements(inchi_to_inchikey, return_dtype=pl.Utf8)
        .alias("inchikey")
    )


def with_clean_string_representations(df: DFLike) -> DFLike:
    """Variant of with_clean_string_representations that uses a join on unique mols
    to avoid rerunning similar mols. Throws away all invalid mols"""
    unique_smiles_df: pl.DataFrame = _clean_mols_og_smiles_df(
        df.select("smiles").unique().collect()["smiles"]  # ty: ignore
    ).drop_nulls("mol")

    original_col_order = list(df.collect_schema().keys())
    mols: pl.Series = unique_smiles_df["mol"]
    return (
        (
            unique_smiles_df.rename({"smiles": "smiles_"})
            .with_columns(
                # smiles
                _series_mols_to_string_representation(mols, type="smiles"),
                # inchis + auxinfo
                *(
                    _series_mols_to_string_representation(
                        mols, type="inchi_and_auxinfo"
                    )
                    .struct.unnest()
                    .iter_columns()
                ),
            )
            # inchikeys directly from the clean inchis
            .with_columns(
                pl.col("inchi")
                .map_elements(inchi_to_inchikey, return_dtype=pl.Utf8)
                .alias("inchikey")
            )
        )
        .lazy()
        .join(
            df.lazy(),
            left_on="smiles_",
            right_on="smiles",
            how="right",
            maintain_order="right",
            suffix="_old",
        )
        .select(*original_col_order, pl.exclude(original_col_order).exclude("mol"))
    )


def with_chembl_approval_year(
    df: pl.LazyFrame | pl.DataFrame,
    chembl_id_col: str = "chembl_id",
    max_calls: int = 10,
    sleep_time: float = 1.0,
) -> pl.LazyFrame | pl.DataFrame:
    """
    Adds a column 'approval_year' to the DataFrame by fetching approval years from ChEMBL API.
    Uses batching and rate limiting to avoid overloading the API.
    """
    # Extract unique, non-null ChEMBL IDs
    uniques = (
        df.lazy()
        .select(pl.col(chembl_id_col))
        .unique()
        .drop_nulls()
        .collect()[chembl_id_col]  # ty: ignore[not-subscriptable]
        .to_list()
    )

    results = chembls_to_approval_years(
        uniques, max_calls=max_calls, sleep_time=sleep_time
    )
    print(results)
    approval_df = (
        pl.DataFrame(
            results, schema={"chembl_id": pl.String, "approval_year": pl.Int32}
        )
        .rename({"chembl_id": chembl_id_col})
        .lazy()
    )

    return df.lazy().join(approval_df, on=chembl_id_col, how="left")


def with_cluster_idx(
    df: DFLike,
    descriptor_col: str = "morgan",
    method="bitbirch",
    alias: str | None = None,
    **kwargs,
):
    # @TODO: check if it is fine to take all unique descriptors or if it should still happen at the mol level
    descriptors = df.lazy().select(descriptor_col).unique().collect()[descriptor_col]  # ty: ignore
    clusters = cluster(
        descriptors=descriptors.to_numpy(),
        method=method,
        **kwargs,
    )
    res = pl.DataFrame(
        [
            descriptors.alias(descriptor_col),
            pl.Series(clusters.tolist(), dtype=pl.Int64).alias(
                alias if alias else method
            ),
        ]
    ).lazy()
    return df.lazy().join(res, on=descriptor_col, how="left", maintain_order="left")


def with_dimred(
    df: DFLike,
    descriptor_col: str = "morgan",
    method: Literal["umap", "tsne"] = "umap",
    metric: str = "auto",
    alias: str | None = None,
    **kwargs,
) -> DFLike:
    descriptors = df.lazy().select(descriptor_col).unique().collect()[descriptor_col]  # ty: ignore
    dimred = dimensionality_reduction(
        descriptors.to_numpy(),  # ty:ignore  [invalid-argument-type]
        method=method,
        metric=metric,
        **kwargs,
    )
    width = dimred.shape[1]
    res = pl.DataFrame(
        [
            descriptors.alias(descriptor_col),
            pl.Series(
                name=alias if alias else method,
                values=dimred.tolist(),
                dtype=pl.Array(pl.Float32, width),
            ),
        ]
    ).lazy()
    return df.lazy().join(res, on=descriptor_col, how="left", maintain_order="left")


def eager_generate_and_filter_by_protein_similarity(
    df: pl.LazyFrame | pl.DataFrame,
    query_df: pl.LazyFrame,
    method: Literal[
        "alphafold_cosine",
        "esm_cosine",
        "global_alignment_blosum62",
        "identity",
        "lehvenstein",
    ] = "esm_cosine",
    threshold: float = 0.7,
) -> pl.LazyFrame | pl.DataFrame:
    """will materialise both dataframes for protein similarities.
    keeps similarities for easy further filtering"""

    max_sims: pl.LazyFrame = (
        protein_similarity_matrix(df.lazy(), query_df.lazy(), method=method)
        .lazy()
        .select(
            pl.col("sequence"),
            pl.max_horizontal(pl.exclude("sequence")).alias("max"),
        )
    )
    return df.lazy().join(
        max_sims.filter(pl.col("max") >= threshold),
        on="sequence",
        how="semi",  # keep df if in max_sims["sequence"]
        maintain_order="left",
    )


def filter_protein_similarity(
    df: pl.LazyFrame,
    prot_with_avoidome_sims: pl.LazyFrame,
    avoidome_df: pl.LazyFrame | None,
    threshold: float = 0.7,
    subset: Literal[
        "all",
        "ion channels",
        "G-protein coupled receptors",
        "nuclear hormone receptors",
        "transporters",
        "enzymes",
        "kinases",
        "first",  # for testing mostly
    ] = "all",
    avoidome_sim_col="avoidome_esm_mean_cosine",
) -> pl.LazyFrame:
    if subset == "first":
        idx = [0]
    elif subset != "all":
        assert avoidome_df is not None, "subset != 'all': avoidome_df must be provided"
        select_ = "" if subset == "all" else subset
        idx = avoidome_df.collect().filter(  # ty: ignore[unresolved-attribute]
            pl.col("type").cast(pl.String).str.contains(select_)
        )["index"]
    else:
        idx = list(
            range(prot_with_avoidome_sims.collect_schema()[avoidome_sim_col].shape[0])  # ty:ignore
        )
    return df.join(
        prot_with_avoidome_sims.filter(
            pl.col(avoidome_sim_col)
            .cast(pl.List)  # workaround for until polars issue #26008 is solved
            .list.gather(idx)
            .list.max()
            >= threshold
        ),
        on="target_id",
        how="semi",  # no info from prot df, only selection
        maintain_order="left",
    )


def with_split_index(
    df,
    X_col: str = "inchikey",
    y_cols: str | list = "pchembl_value_mean",
    cluster_col: str = "cluster",
    method: Literal["tricario", "sklearn"] = "tricario",
    n_splits: int = 5,
    alias: str = "split",
    **kwargs,
):
    if isinstance(y_cols, str):
        y_cols = [y_cols]
    clusters, split_idxs = split(
        df=df,
        X_col=X_col,
        y_cols=y_cols,
        cluster_col=cluster_col,
        n_splits=n_splits,
        method=method,
        **kwargs,
    )
    assert len(clusters) == len(split_idxs)
    mapping = pl.DataFrame(
        [
            pl.Series(name=cluster_col, values=clusters),
            pl.Series(name=alias, values=split_idxs),
        ]
    ).unique()  # to handle the
    return df.lazy().join(
        mapping.lazy(),
        on=cluster_col,
        how="left",
    )


def __conditional_rename(
    df: pl.DataFrame | pl.LazyFrame, dict_: dict[str, str] | None
) -> pl.DataFrame | pl.LazyFrame:
    if not dict_:
        return df
    dict_ = {k: v for k, v in dict_.items() if k in df.collect_schema()}
    return df.rename(dict_)


def with_query_sim_array(
    df: pl.DataFrame | pl.LazyFrame,
    query_df: pl.DataFrame | pl.LazyFrame,
    method: str,
    alias: str = "protein_similarity",
    use_col_as: dict[str, str] | None = None,
) -> pl.LazyFrame:
    """compresses the similarity matrix df to pl.Array, keeping the order

    Args:
        df (pl.DataFrame | pl.LazyFrame): dataframe with 'sequence' column
        query_df (pl.DataFrame | pl.LazyFrame): dataframe with 'sequence' column
        method (str): method for protein similarity
        alias (str, optional): name for the similarity array column. Defaults to "protein_similarity".
        use_col_as (dict[str, str] | None, optional): mapping of column names to use as standard
            names. useful when using embeddings. Example could be `{"esm_max_pooling":"esm"}`.
            Will not raise on missing items and will try to apply on both. In case the columns
            do not exist and embeddings are needed, will just calculate the embeddings from scratch.
            Defaults to None.
    Returns:
        pl.LazyFrame: original df with added similarity array column

    """

    idx_sequence = query_df.select(["index", "sequence"]).sort("index").lazy().collect()

    matrix = protein_similarity_matrix(
        df.unique("sequence").pipe(__conditional_rename, use_col_as).lazy(),
        query_df.pipe(__conditional_rename, use_col_as).lazy(),
        method=method,
    )
    n = matrix.drop("sequence").width
    matrix = matrix.select(
        pl.col("sequence"),
        (
            pl.concat_list(pl.col(idx_sequence.get_column("sequence")))  # ty: ignore[unresolved-attribute]
            .list.to_array(n)
            .alias(alias)
        ),
    )
    return df.lazy().join(matrix.lazy(), on="sequence", how="left")


# ====== utils =======


def task_readout_per_activity(
    flat_df: pl.DataFrame | pl.LazyFrame,
    readout_fmt: str = "^type_.*$",
    group_by="activity_id",
) -> pl.DataFrame | pl.LazyFrame:
    """calculates the sum of readouts per activity. useful for stratification for splitting.
    Args:
        - flat_df (pl.DataFrame | pl.LazyFrame): dataframe with single datapoints
            that have one-hot-encodings for each activity column
        - readout_fmt (str): regex expression (or single string) for selecting the readout columns
        - group_by (str): column name for grouping by, e.g. `activity_id`,`target_id`,`smiles`

    Returns:
        - pl.LazyFrame with columns [group_by, readout_fmt] where readout_fmt
        is the sum of the readouts for each group_by unit.

    Example usage:
        `pl.scan_parquet('papyrus.parquet').pipe(semicolon_to_flat).pipe(task_readout_per_activity)`
    """
    return (
        flat_df.select(pl.col(readout_fmt), group_by)
        .group_by(group_by)
        .agg(pl.col(readout_fmt).cast(pl.Int8).sum())
    )


def task_protein_per_mol(
    flat_df, target_col="target_id", mol_col="inchi"
) -> pl.LazyFrame:
    return flat_df.select(target_col, mol_col).pivot(
        on=target_col,
        on_columns=flat_df.select(target_col).unique().collect().to_series(),
        index=mol_col,
        values=target_col,
        aggregate_function="len",
    )


def filter_is_in(lf, lf_series_filter):
    """join-based filter is_in for a lazyframe.
    Args:
        lf (pl.LazyFrame): lazyframe to filter
        lf_series_filter (pl.LazyFrame): lf for which values to keep. if not single-column, will use all columns. Usually only single-columned LF.

    Note: could be used with pl.DataFrames too but I generally avoid joins with eager df's due to memory concerns
    """
    return lf.lazy().join(
        lf_series_filter.unique().lazy(),
        on=lf_series_filter.collect_schema().names(),  # all cols
        how="semi",  # keep left if in right
        maintain_order="left",
    )


# ======== prep


def with_chemprop_datapoints_from_mols(
    df: pl.DataFrame,
    X_d: str | None,
    mol_col: str = "mol",
    y: str = "pchembl_value_mean",
) -> pl.DataFrame:
    assert X_d is None or X_d in df.collect_schema()
    assert y is None or y in df.collect_schema()
    df = df.drop("index", strict=False).with_row_index().sort("index")
    mols = df.lazy().select(mol_col).collect().get_column(mol_col).to_list()  # ty: ignore[unresolved-attribute]
    prot_descs = (
        df.lazy().select(X_d).collect().get_column(X_d).to_numpy() if X_d else None  # ty: ignore[unresolved-attribute]
    )
    ys = df.lazy().select(y).collect().get_column(y).to_list() if y else None  # ty: ignore[unresolved-attribute]
    return df.sort("index").with_columns(
        pl.Series(
            create_molecule_datapoints(mols, prot_descs, ys), dtype=pl.Object
        ).alias("chemprop_datapoints")
    )


def with_chemprop_datapoints_from_inchi(
    df: pl.DataFrame,
    X_d: str | None,
    y: str = "pchembl_value_mean",
    parallel_mols: bool = True,
) -> pl.DataFrame:
    assert X_d is None or X_d in df.collect_schema()
    assert y is None or y in df.collect_schema()
    df = df.pipe(with_mols_from_inchi, parallel=parallel_mols)  # ty:ignore[invalid-assignment]
    return with_chemprop_datapoints_from_mols(
        df, X_d, mol_col="mol", y="pchembl_value_mean"
    ).drop("mol")


def with_chemprop_avoidome_datapoints(
    df: pl.DataFrame,
    avoidome: pl.DataFrame,
    prot_desc: str,
    parallel_mols: bool = False,
) -> pl.DataFrame:
    assert prot_desc in avoidome.collect_schema()
    df = (  # ty: ignore[invalid-assignment]
        df.join(avoidome.select("accession", "sequence", prot_desc), how="cross")
        .with_row_index()
        .pipe(with_mols_from_inchi, parallel=parallel_mols)
        .sort("index")
        .collect()  # ty: ignore[ unresolved-attribute]
    )
    mols = df["mol"]
    prot_descs = df[prot_desc].to_numpy() if prot_desc else None
    return (
        df.sort("index")
        .with_columns(
            pl.Series(
                create_molecule_datapoints(mols, prot_descs, ys=None), dtype=pl.Object
            ).alias("chemprop_datapoints")
        )
        .drop("mol")
    )



import logging as lg
from functools import lru_cache, partial, reduce
from pathlib import Path
from typing import Callable, Iterable

from rdkit.Chem import (
    Kekulize,
    Mol,
    MolFromInchi,
    MolFromSmiles,
    MolToInchi,
    MolToInchiAndAuxInfo,
    MolToInchiKey,
    RemoveStereochemistry,
)
from rdkit.Chem.inchi import InchiToInchiKey
from rdkit.Chem.rdmolfiles import MolToSmiles

from fairfetched.standardisation.optional import papyrus_standardise
from fairfetched.utils._track import track

# def _fuse_pipeline(fns: tuple[Callable, ...]) -> Callable:
#     return reduce(lambda f, g: lambda m: (r := f(m)) and g(r), fns)


def fuse_pipeline(
    *fns: Callable[[Mol | str | None], Mol | tuple[str, str, str] | None],
) -> Callable:
    """Combines functions on mols into one function. Technically made for mol-to-mol|None functions,
    but could be started and finished with Any->Mol and Mol->Any functions, respectively

    Example usage:
        ```
        standardise = _fuse_pipeline(remove_stereochemistry,validate_inchi,standardise_papyrus)
        ```
    """
    return reduce(lambda f, g: lambda m: (r := f(m)) and g(r), tuple(fns))


def _safe_smiles_to_mol(smiles) -> Mol | None:
    try:
        return MolFromSmiles(smiles)
    except Exception as e:
        lg.error(f"Failed on `MolFromSmiles` conversion: {smiles} with exception {e}")
        return None


def _safe_inchi_to_mol(inchi: str) -> Mol | None:
    try:
        return MolFromInchi(inchi)
    except Exception as e:
        lg.error(e)
        return None


def remove_stereochemistry(mol) -> Mol | None:
    try:
        RemoveStereochemistry(mol)
    except Exception as e:
        lg.error(f"Failed in RemoveStereochemistry(mol) with error {e}")
        return None
    return mol


def validate_inchi(mol) -> Mol | None:
    try:
        inchi = MolToInchi(mol)
        mol = MolFromInchi(inchi)
        if not isinstance(mol, Mol):
            lg.error(f"Failed on InChI conversion: {inchi}")
            return None
        return mol
    finally:
        return None


def standardise_papyrus(mol: Mol | None, **kwargs) -> Mol | None:
    if not mol:
        return None
    try:
        mol, message = papyrus_standardise(
            mol, return_type=True, raise_error=False, **kwargs
        )
        if not mol:
            lg.error(f"Papyrus-type standardisation did not return a mol: {message}")
            return None
        return mol
    except Exception as e:
        lg.error(f"failed papyrus standardisation with exception: {e}")
        return None


def mol_to_strings(mol: Mol | None) -> tuple[str, str, str, str] | None:
    """returns a tuple of kekulised smiles, inchi, inchiauxinfo, and inchikeys"""
    if not Mol:
        return None
    return MolToSmiles(mol), *MolToInchiAndAuxInfo(mol), MolToInchiKey(mol)


standardise_aeffect = fuse_pipeline(
    remove_stereochemistry,
    partial(
        standardise_papyrus,
        filter_mixtures=True,  # @TODO: check if this is desired
        filter_inorganic=False,
        filter_non_small_molecule=False,
        return_type=True,
        raise_error=False,
    ),
    validate_inchi,
)


# def _smiles_to_clean_mol(smiles: str) -> Mol | None:
#     """Returns cleaned molecules from SMILES string or None upon failure or invalid molecules"""
#     assert isinstance(smiles, str)
#     if not smiles:
#         lg.warning("empty smiles")
#         return None
#     mol = MolFromSmiles(smiles)
#     if not isinstance(mol, Mol):
#         lg.debug("Failed on initial `MolFromSmiles` conversion")
#         return None
#     try:
#         RemoveStereochemistry(mol)
#     except Exception as e:
#         lg.error(f"Failed in RemoveStereochemistry(mol) with error {e}")
#         return None
#     # @TODO: check if stereochemistry removal should be before or after standardisation
#     if not mol:
#         lg.error(f"{smiles} could not be converted to a mol")
#         return None
#     try:
#         mol, message = papyrus_standardise(
#             mol,
#             filter_mixtures=True,  # @TODO: check if this is desired
#             filter_inorganic=False,
#             filter_non_small_molecule=False,
#             return_type=True,
#             raise_error=False,
#         )
#         if not mol:
#             lg.warning(
#                 f"{smiles} could not be cleaned during Papyrus-type standardisation: {message}"
#             )
#             return None
#         mol = MolFromInchi(MolToInchi(mol))
#         if not mol:
#             lg.warning(f"{smiles} did not give valid inchi: {message}")
#         return mol
#     except Exception as e:
#         lg.error(
#             f"{smiles} could not be cleaned in papyrus-style standardsation with exception: {e}"
#         )
#         return None
#     return None


def inchis_to_mols(
    inchis: Iterable[str], cache: Path | None = None
) -> list[Mol | None]:
    """Convert InChIs to RDKit Mol objects"""
    return [
        _safe_inchi_to_mol(i) for i in track(inchis, desc="Converting InChI to mol")
    ]


def _mol_to_kekulised_smiles(mol: Mol) -> str:
    """Returns a SMILES string suitable for re-loading the molecule later.
    Removes stereochemistry.
    """
    try:
        Kekulize(mol, clearAromaticFlags=True)
    except Exception as e:
        lg.debug(f"Kekulization failed with error {e}")
        return MolToSmiles(mol, isomericSmiles=False)
    return MolToSmiles(mol, isomericSmiles=False)


def mols_to_kekulised_smiles(mols: Iterable[Mol]) -> list[str]:
    """Returns kekulised SMILES strings suitable for re-loading the molecule later.
    Removes stereochemistry.
    """
    return [
        _mol_to_kekulised_smiles(m)
        for m in track(mols, desc="converting mols to kekulised smiles")
    ]


def _mol_to_inchi_and_auxinfo(mol: Mol) -> tuple[str, str]:
    return MolToInchiAndAuxInfo(mol)


def _mol_to_inchi(mol: Mol) -> str:
    return MolToInchiAndAuxInfo(mol)[0]  # to maintain consistency


def mols_to_inchis(mols: Iterable[Mol]) -> list[str]:
    return [_mol_to_inchi(m) for m in track(mols, desc="converting mols to inchis")]


def mols_to_inchis_and_auxinfo(mols: Iterable[Mol]) -> list[tuple[str, str]]:
    return [
        _mol_to_inchi_and_auxinfo(m)
        for m in track(mols, desc="converting mols to inchis with auxinfo")
    ]


def inchis_to_inchikeys(inchis: Iterable[str]) -> list[str]:
    return [
        InchiToInchiKey(i) for i in track(inchis, desc="converting inchis to inchikeys")
    ]


def mols_to_inchikeys(mols: Iterable[Mol]) -> list[str]:
    return [
        InchiToInchiKey(MolToInchiAndAuxInfo(m)[0])
        for m in track(mols, desc="converting mols to inchikeys")
    ]


mol_to_kekulized_smiles = _mol_to_kekulised_smiles  # American spelling alias
inchi_to_inchikey = InchiToInchiKey  # convenience alias



import os
from concurrent.futures import ProcessPoolExecutor
from functools import partial
from multiprocessing import Pool
from typing import Callable, Iterable, Literal

import polars as pl
from rdkit.Chem import Mol

from fairfetched.utils._track import track
from fairfetched.utils.polars import apply_to_unique, map_batches_pooled

from .compound_fns import (
    _remove_stereo_papyrus_standardise_check_inchi,
    _safe_inchi_to_mol,
    _safe_smiles_to_mol,
    inchi_to_inchikey,
    mols_to_inchikeys,
    mols_to_inchis,
    mols_to_inchis_and_auxinfo,
    mols_to_kekulised_smiles,
    standardised_nostereo_to_smiles_inchi_aux_inchikey,
)


def apply_to_unique_strings(
    lf: pl.LazyFrame,
    function: Callable[[str], Mol | None] = _safe_inchi_to_mol,
    from_col: str = "inchi",
    to_col: str = "mol",
    parallel=True,
    chunks_per_worker: int = 8,
) -> pl.LazyFrame:
    """
    for ~1M mols: 1h07m for fully linear one, 15m for this impl. (now actually <1m)
    chunksize is calculated as max(1, len(unique_molstrings) // (workers * chunks_per_worker))
    so for smaller chunks (good for larger operations), you would need more chunks per worker (800 or so)
    """
    if parallel:
        unique_molstrings = (
            lf.select(pl.col(from_col)).unique().collect().get_column(from_col)  # ty: ignore[unresolved-attribute]
        )
        workers = round(os.cpu_count() or 1 * 0.8)
        chunksize = max(1, len(unique_molstrings) // (workers * chunks_per_worker))
        with ProcessPoolExecutor(max_workers=workers) as p:
            res = track(
                p.map(function, unique_molstrings, chunksize=chunksize),
                total=len(unique_molstrings),
            )

            mols_df = pl.DataFrame(
                [
                    pl.Series(unique_molstrings, dtype=pl.Utf8).alias(from_col),
                    pl.Series(list(res), dtype=pl.Object).alias(to_col),
                ]
            )
        return lf.join(
            mols_df.lazy(),
            on=from_col,
            how="left",
            maintain_order="left",
        )

    return lf.join(
        lf.select(pl.col(from_col))
        .unique()
        .with_columns(
            pl.col(from_col)
            .map_elements(function, return_dtype=pl.Object)
            .alias(to_col)
        ),
        on=from_col,
        how="left",
        maintain_order="left",
    )


def with_cleaned_mol_descriptors_struct(
    lf: pl.LazyFrame,
    descriptor_to_descriptors_func: Callable[[str], tuple[str | None, ...]],
    from_col: str,
    to_col: str,
    return_dtype: pl.DataTypeExpr,
    parallel=True,
    **kwargs,
) -> pl.LazyFrame:
    return lf.pipe(
        apply_to_unique,  # creates unique list, iterates over pooled,
        descriptor_to_descriptors_func,
        from_col=from_col,
        to_col=to_col,
        return_dtype=return_dtype,
        parallel=True,
    )


def with_cleaned_mol_descriptors(
    lf: pl.LazyFrame,
    smiles_col: str = "smiles",
    **kwargs,
) -> pl.LazyFrame:
    """cleans and standardises mols by:
    1) removing stereochemistry
    2) running papyrus standardisation
    3) validating inchis
    returns None if invalid, but does not drop
    """
    fields = ["smiles", "inchi", "inchi_auxinfo", "inchikey"]
    to_col = "descriptors_struct"
    lf = lf.pipe(
        with_cleaned_mol_descriptors_struct,
        descriptor_to_descriptors_func=standardised_nostereo_to_smiles_inchi_aux_inchikey,
        from_col=smiles_col,
        to_col=to_col,
        return_dtype=pl.Struct(
            {k: pl.Utf8 for k in ["smiles", "inchi", "inchi_auxinfo", "inchikey"]}
        ),
        **kwargs,
    )
    return (
        lf.cast({to_col: pl.Struct({k: pl.Utf8 for k in fields})})
        .rename({k: f"{k}_original" for k in fields if k in lf.collect_schema()})
        .with_columns(pl.col(to_col).struct.unnest())
        .drop(to_col)
    )


def with_mols_from_inchi_batched(lf, alias="mol", parallel=False):
    function = partial(
        map_batches_pooled,
        fn=_safe_inchi_to_mol,
        return_dtype=pl.Object,
        parallel=parallel,
        desc="loading mols",  # used only if parallel => tracks progress
    )
    return lf.with_columns(
        pl.col("inchi")
        .map_batches(function, return_dtype=pl.Object, is_elementwise=not parallel)
        .alias(alias)
    )


def with_mols_from_smiles_batched(lf, alias="mol", parallel=False):
    function = partial(
        map_batches_pooled,
        fn=_safe_smiles_to_mol,
        return_dtype=pl.Object,
        parallel=parallel,
        desc="loading mols",  # used only if parallel => tracks progress
    )
    return lf.with_columns(
        pl.col("smiles")
        .map_batches(function, return_dtype=pl.Object, is_elementwise=not parallel)
        .alias(alias)
    )


def with_mols_from_inchi_pool(
    lf: pl.LazyFrame, alias="mol", parallel=True
) -> pl.LazyFrame:
    """
    for ~1M mols: 1h07m for fully linear one, 15m for this impl.
    """
    if parallel:
        unique_inchis = (
            lf.select(pl.col("inchi")).unique().collect().get_column("inchi")  # ty: ignore[unresolved-attribute]
        )
        with Pool() as p:
            res = p.map(_safe_inchi_to_mol, unique_inchis)
        mols_df = pl.DataFrame(
            [
                pl.Series(unique_inchis, dtype=pl.Utf8).alias("inchi"),
                pl.Series(res, dtype=pl.Object).alias(alias),
            ]
        )
        return lf.join(
            mols_df.lazy(),
            on="inchi",
            how="left",
            maintain_order="left",
        )
    return lf.join(
        lf.select(pl.col("inchi"))
        .unique()
        .with_columns(
            pl.col("inchi")
            .map_elements(_safe_inchi_to_mol, return_dtype=pl.Object)
            .alias(alias)
        ),
        on="inchi",
        how="left",
        maintain_order="left",
    )


def with_mols_from_inchi_thread(
    lf: pl.LazyFrame,
    alias="mol",
    parallel=True,
    join=True,
) -> pl.LazyFrame:
    """
    Returns a new DataFrame with a column of RDKit Mol objects from InChI strings.
    """
    if parallel:
        if join:
            inchis = (
                lf.select(pl.col("inchi")).unique().collect().get_column("inchi")  # ty: ignore[unresolved-attribute]
            )
        else:
            lf = lf.sort("inchi")
            inchis = lf.select("inchi").collect().get_column("inchi")  # ty: ignore[unresolved-attribute]
            inchis = lf.select("inchi").collect().get_column("inchi")  # ty: ignore[unresolved-attribute]
        workers = round((os.cpu_count() or 1) * 0.9)
        chunksize = max(1, len(inchis) // (workers * 4))
        print("n_workers=", workers)
        print("chunksize=", chunksize)
        with ProcessPoolExecutor(max_workers=workers) as executor:
            res = list(
                executor.map(
                    _safe_inchi_to_mol,
                    track(inchis, total=len(inchis), desc="loading mols"),
                    chunksize=chunksize,
                )
            )
            mol_series = pl.Series(res, dtype=pl.Object).alias(alias)
        if join:
            mols_df = pl.DataFrame(
                [pl.Series(inchis, dtype=pl.Utf8).alias("inchi"), mol_series]
            )
            return lf.join(
                mols_df.lazy(),
                on="inchi",
                how="left",
                maintain_order="left",
            )
        else:
            lf.with_columns(mol_series)
    return lf.join(
        lf.select(pl.col("inchi"))
        .unique()
        .with_columns(
            pl.col("inchi")
            .map_elements(_safe_inchi_to_mol, return_dtype=pl.Object)
            .alias(alias)
        ),
        on="inchi",
        how="left",
        maintain_order="left",
    )


def with_clean_mols_from_smiles(df: pl.LazyFrame, alias="mol") -> pl.LazyFrame:
    return df.with_columns(
        pl.col("smiles")
        .map_elements(
            _remove_stereo_papyrus_standardise_check_inchi,
            return_dtype=pl.Object,
            strategy="threading",
        )
        .alias("mol")
    )


def _smiles_and_cleaned_mols_df(smiles: Iterable[str]) -> pl.DataFrame:
    with Pool() as p:
        res = p.map(_remove_stereo_papyrus_standardise_check_inchi, smiles)
    return pl.DataFrame(
        [
            pl.Series(smiles, dtype=pl.Utf8).alias("smiles"),
            pl.Series(res, dtype=pl.Object).alias("mol"),
        ]
    )


def _series_mols_to_string_representation(
    mols: pl.Series,
    type: Literal["smiles", "inchi", "inchi_and_auxinfo", "inchikey"] = "smiles",
) -> pl.Series:
    match type:
        case "smiles":
            func = mols_to_kekulised_smiles
            dtype = pl.Utf8
        case "inchi":
            func = mols_to_inchis
            dtype = pl.Utf8
        case "inchi_and_auxinfo":
            func = mols_to_inchis_and_auxinfo
            dtype = pl.Struct({"inchi": pl.Utf8, "inchi_auxinfo": pl.Utf8})
        case "inchikey":
            func = mols_to_inchikeys
            dtype = pl.Utf8
        case _:
            raise NotImplementedError
    return pl.Series(name=type, values=func(mols), dtype=dtype)


# def _unique_with_clean_string_representations(df: pl.LazyFrame) -> pl.LazyFrame:
#     """cleans and standardises molecules, replacing original string
#     representations (which are tagged _papyrus) and adding new ones
#     without the suffix. This is to ensure that all string representations
#     are consistent and can be used for downstream tasks without confusion."""
#     df = df.pipe(with_clean_mols_from_smiles).rename(
#         {k: f"{k}_papyrus" for k in ["smiles", "inchi", "inchi_auxinfo", "inchikey"]},
#         strict=False,
#     )
#     mols = df.lazy().select("mol").collect().get_column("mol")
#     return df.with_columns(
#         _series_mols_to_string_representation(mols, type="smiles"),
#         _series_mols_to_string_representation(
#             mols, type="inchi_and_auxinfo"
#         ).struct.unnest(),
#     ).with_columns(
#         pl.col("inchi")
#         .map_elements(inchi_to_inchikey, return_dtype=pl.Utf8)
#         .alias("inchikey")
#     )


def with_clean_string_representations(df: pl.LazyFrame) -> pl.LazyFrame:
    """Variant of with_clean_string_representations that uses a join on unique mols
    to avoid rerunning similar mols. Throws away all invalid mols"""
    unique_smiles_df: pl.DataFrame = _smiles_and_cleaned_mols_df(
        df.select("smiles").unique().collect()["smiles"]  # ty: ignore
    ).drop_nulls("mol")

    original_col_order = list(df.collect_schema().keys())
    mols: pl.Series = unique_smiles_df["mol"]
    return (
        (
            unique_smiles_df.rename({"smiles": "smiles_"})
            .with_columns(
                # sm iles
                _series_mols_to_string_representation(mols, type="smiles"),
                # inchis + auxinfo
                *(
                    _series_mols_to_string_representation(
                        mols, type="inchi_and_auxinfo"
                    )
                    .struct.unnest()
                    .iter_columns()
                ),
            )
            # inchikeys directly from the clean inchis
            .with_columns(
                pl.col("inchi")
                .map_elements(inchi_to_inchikey, return_dtype=pl.Utf8)
                .alias("inchikey")
            )
        )
        .lazy()
        .join(
            df.lazy(),
            left_on="smiles_",
            right_on="smiles",
            how="right",
            maintain_order="right",
            suffix="_old",
        )
        .select(*original_col_order, pl.exclude(original_col_order).exclude("mol"))
    )


def __conditional_rename(
    df: pl.DataFrame | pl.LazyFrame, dict_: dict[str, str] | None
) -> pl.DataFrame | pl.LazyFrame:
    if not dict_:
        return df
    dict_ = {k: v for k, v in dict_.items() if k in df.collect_schema()}
    return df.rename(dict_)


# ====== utils =======


def task_readout_per_activity(
    flat_df: pl.DataFrame | pl.LazyFrame,
    readout_fmt: str = "^type_.*$",
    group_by="activity_id",
) -> pl.DataFrame | pl.LazyFrame:
    """calculates the sum of readouts per activity. useful for stratification for splitting.
    Args:
        - flat_df (pl.DataFrame | pl.LazyFrame): dataframe with single datapoints
            that have one-hot-encodings for each activity column
        - readout_fmt (str): regex expression (or single string) for selecting the readout columns
        - group_by (str): column name for grouping by, e.g. `activity_id`,`target_id`,`smiles`

    Returns:
        - pl.LazyFrame with columns [group_by, readout_fmt] where readout_fmt
        is the sum of the readouts for each group_by unit.

    Example usage:
        `pl.scan_parquet('papyrus.parquet').pipe(semicolon_to_flat).pipe(task_readout_per_activity)`
    """
    return (
        flat_df.select(pl.col(readout_fmt), group_by)
        .group_by(group_by)
        .agg(pl.col(readout_fmt).cast(pl.Int8).sum())
    )


def task_protein_per_mol(
    flat_df, target_col="target_id", mol_col="inchi"
) -> pl.LazyFrame:
    return flat_df.select(target_col, mol_col).pivot(
        on=target_col,
        on_columns=flat_df.select(target_col).unique().collect().to_series(),
        index=mol_col,
        values=target_col,
        aggregate_function="len",
    )


def filter_is_in(lf, lf_series_filter):
    """join-based filter is_in for a lazyframe.
    Args:
        lf (pl.LazyFrame): lazyframe to filter
        lf_series_filter (pl.LazyFrame): lf for which values to keep. if not single-column, will use all columns. Usually only single-columned LF.

    Note: could be used with pl.DataFrames too but I generally avoid joins with eager df's due to memory concerns
    """
    return lf.lazy().join(
        lf_series_filter.unique().lazy(),
        on=lf_series_filter.collect_schema().names(),  # all cols
        how="semi",  # keep left if in right
        maintain_order="left",
    )



from sqlalchemy import create_engine, inspect


def print_sqlite_schema(db_path: str) -> None:
    import sqlite3

    with sqlite3.connect(db_path) as conn:
        cur = conn.cursor()

        cur.execute("""
        SELECT name
        FROM sqlite_master
        WHERE type='table' AND name NOT LIKE 'sqlite_%'
        ORDER BY name;
        """)
        tables = [r[0] for r in cur.fetchall()]

        for table in tables:
            print(f"\n{table}")
            cur.execute(f"PRAGMA table_info('{table}')")
            for cid, name, coltype, notnull, default, pk in cur.fetchall():
                nn = " NOT NULL" if notnull else ""
                pkf = " PK" if pk else ""
                dflt = f" DEFAULT {default}" if default is not None else ""
                print(f"  {name} ({coltype}){nn}{pkf}{dflt}")


def load_schema(db_path: str):
    engine = create_engine(f"sqlite:///{db_path}")
    inspector = inspect(engine)

    schema = {}

    for table in inspector.get_table_names():
        cols = inspector.get_columns(table)
        pk_info = inspector.get_pk_constraint(table)
        pks = set(pk_info.get("constrained_columns") or [])
        fks = inspector.get_foreign_keys(table)

        fk_cols = {
            fk["constrained_columns"][0] for fk in fks if fk.get("constrained_columns")
        }

        schema[table] = {
            "columns": [
                {
                    "name": c["name"],
                    "type": str(c["type"]).split("(")[0],
                    "pk": c["name"] in pks,
                    "fk": c["name"] in fk_cols,
                }
                for c in cols
            ],
            "foreign_keys": [
                {
                    "column": fk["constrained_columns"][0],
                    "target_table": fk["referred_table"],
                }
                for fk in fks
                if fk.get("constrained_columns") and fk.get("referred_table")
            ],
        }

    return schema


def schema_to_mermaid(schema):
    lines = ["erDiagram"]

    # Tables
    for table, data in schema.items():
        lines.append(f"    {table} {{")
        for col in data["columns"]:
            if col["pk"]:
                tag = " PK"
            elif col["fk"]:
                tag = " FK"
            else:
                tag = ""
            lines.append(f"        {col['type']} {col['name']}{tag}")
        lines.append("    }")

    # Relationships
    for table, data in schema.items():
        for fk in data["foreign_keys"]:
            lines.append(f"    {table} }}|--|| {fk['target_table']} : {fk['column']}")

    return "\n".join(lines)


def db_to_mermaid(db_path: str):
    return schema_to_mermaid(load_schema(db_path))



from pathlib import Path
from typing import Any, NotRequired, Protocol, Tuple, TypedDict

from polars import LazyFrame


class ComposedLFDict(TypedDict):
    """a dictionary with different lazyframes composed from raw cheminformatics
    databases. requires at minimum a bioactivity and compounds dataframe"""

    bioactivity: LazyFrame
    compounds: LazyFrame
    bioactivity_nostereochemistry: NotRequired[LazyFrame]
    full: NotRequired[LazyFrame]
    full_nostereochemistry: NotRequired[LazyFrame]
    proteins: NotRequired[LazyFrame]
    components: NotRequired[LazyFrame]


class DatasetGetModule(Protocol):
    """Protocol for the logic required in the get.papyrus, get.chembl files

    For referencing use in the general _Base API of get.api.
    """

    __name__: str
    __file__: str

    def available_versions(self) -> Tuple[str]:
        """returns available versions of database"""
        ...

    def latest(self) -> str:
        """returns latest version of database"""
        ...

    def get_sources(self, version: str) -> dict[str, str]:
        """returns a dict of URLs of the sources of database"""
        ...

    def ensure_raw(
        self, version: str, raw_dir: Path | str | Any | None = None
    ) -> dict[str, Path]:
        """downloads raw files for specified version from database sources
        to cache dir if not yet present, and returns a dictionary of the source
        names and the paths to their raw files."""
        ...

    def ensure_consolidated(
        self,
        raw_paths: dict[str, Path],
        consolidated_dir: Path | str | Any | None = None,
    ) -> dict[str, Path]:
        """from dictionary of raw files, makes sure that all files are in parquet
        format to allow lazy loading with pl.scan. assigns datatypes and
        consolidates null values in case of ambiguous formats such as csv"""
        ...

    def clean(self, paths: dict[str, Path]) -> dict[str, LazyFrame]:
        """does rest of standardisation, such as column renaming etc"""
        ...

    def compose(self, lfs: dict[str, LazyFrame]) -> ComposedLFDict:
        """composes into few, comprehensive lazyframes through joins and
        pivots."""
        ...


#     "ComposedDict",
#     {
#         "bioactivity": pl.LazyFrame,
#         "compounds": pl.LazyFrame,
#         "proteins": NotRequired[pl.LazyFrame],
#     },
#     total=False,
# )



"""Integration tests for the chembl module.

These tests verify the end-to-end workflow of downloading, extracting,
and processing ChEMBL data, using temporary directories that are cleaned
up after each test.
"""

import shutil
import sqlite3
import tarfile
import tempfile
from pathlib import Path
from unittest.mock import Mock, patch

import pytest

from fairfetched.utils.ensure import ensure_url
from fairfetched.utils.files import ensure_untarred_sqlite
from fairfetched.utils.polars import ensure_sqlite_db_to_parquets


@pytest.fixture
def temp_base_dir():
    """Create a temporary base directory for all test data."""
    temp_path = Path(tempfile.mkdtemp())
    yield temp_path
    # Cleanup after test
    if temp_path.exists():
        shutil.rmtree(temp_path)


@pytest.fixture
def sample_chembl_database():
    """Create a mock ChEMBL-like SQLite database with sample tables."""
    with tempfile.NamedTemporaryFile(suffix=".db", delete=False) as tf:
        temp_db = Path(tf.name)
    conn = sqlite3.connect(temp_db)
    cursor = conn.cursor()

    # Create sample tables mimicking ChEMBL structure
    cursor.execute("""
        CREATE TABLE molecule_dictionary (
            molregno INTEGER PRIMARY KEY,
            chembl_id TEXT UNIQUE NOT NULL,
            pref_name TEXT
        )
    """)

    cursor.execute("""
        CREATE TABLE compound_properties (
            molregno INTEGER PRIMARY KEY,
            mw_freebase REAL,
            alogp REAL
        )
    """)

    cursor.execute("""
        CREATE TABLE activities (
            activity_id INTEGER PRIMARY KEY,
            assay_id INTEGER,
            molregno INTEGER,
            standard_value REAL,
            standard_units TEXT
        )
    """)

    cursor.execute("""
        CREATE TABLE assays (
            assay_id INTEGER PRIMARY KEY,
            assay_type TEXT,
            description TEXT
        )
    """)

    # Insert sample data
    cursor.execute(
        "INSERT INTO molecule_dictionary VALUES (?, ?, ?)",
        (1, "CHEMBL1", "Aspirin"),
    )
    cursor.execute(
        "INSERT INTO molecule_dictionary VALUES (?, ?, ?)",
        (2, "CHEMBL2", "Ibuprofen"),
    )

    cursor.execute(
        "INSERT INTO compound_properties VALUES (?, ?, ?)",
        (1, 180.16, 1.19),
    )
    cursor.execute(
        "INSERT INTO compound_properties VALUES (?, ?, ?)",
        (2, 206.28, 3.97),
    )

    cursor.execute(
        "INSERT INTO activities VALUES (?, ?, ?, ?, ?)",
        (1, 1, 1, 5.2, "nM"),
    )
    cursor.execute(
        "INSERT INTO activities VALUES (?, ?, ?, ?, ?)",
        (2, 2, 2, 3.1, "nM"),
    )

    cursor.execute(
        "INSERT INTO assays VALUES (?, ?, ?)",
        (1, "B", "Binding assay"),
    )
    cursor.execute(
        "INSERT INTO assays VALUES (?, ?, ?)",
        (2, "F", "Functional assay"),
    )

    conn.commit()
    conn.close()

    yield temp_db

    # Cleanup
    if temp_db.exists():
        temp_db.unlink()


@pytest.fixture
def chembl_tar_gz(sample_chembl_database, temp_base_dir):
    """Create a tar.gz archive containing the mock ChEMBL database."""
    tar_gz_path = temp_base_dir / "chembl_36_sqlite.tar.gz"

    with tarfile.open(tar_gz_path, mode="w:gz") as tar:
        tar.add(sample_chembl_database, arcname="chembl_36.db")

    return tar_gz_path


def test_ensure_url_downloads_tar_gz(chembl_tar_gz, temp_base_dir):
    """Test downloading a tar.gz file (mocked)."""
    download_dir = temp_base_dir / "raw"
    download_dir.mkdir(parents=True, exist_ok=True)

    # Read the actual tar.gz content
    tar_gz_content = chembl_tar_gz.read_bytes()

    mock_resp = Mock()
    mock_resp.read = Mock(side_effect=[tar_gz_content, b""])
    mock_resp.getheader = Mock(return_value=str(len(tar_gz_content)))
    mock_resp.__enter__ = Mock(return_value=mock_resp)
    mock_resp.__exit__ = Mock(return_value=None)

    with patch("urllib.request.urlopen", return_value=mock_resp):
        result = ensure_url(
            "https://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/releases/chembl_36/chembl_36_sqlite.tar.gz",
            download_dir / "chembl_36_sqlite.tar.gz",
        )

    # Verify the file was downloaded
    assert result.exists()
    assert result.suffix == ".gz"
    assert result.stat().st_size > 0


def test_extract_and_convert_to_parquets(chembl_tar_gz, temp_base_dir):
    """Test extracting tar.gz and converting to parquet files."""
    # Extract the database
    extracted_db = ensure_untarred_sqlite(chembl_tar_gz)
    assert extracted_db.exists()

    # Convert to parquets
    parquet_dir = temp_base_dir / "clean"
    parquet_paths = ensure_sqlite_db_to_parquets(extracted_db, cache_dir=parquet_dir)

    # Verify parquet files were created
    assert len(parquet_paths) > 0
    assert parquet_dir.exists()

    # Verify expected tables
    expected_tables = [
        "molecule_dictionary",
        "compound_properties",
        "activities",
        "assays",
    ]
    for table_name in expected_tables:
        assert table_name in parquet_paths
        parquet_path = parquet_paths[table_name]
        assert parquet_path.exists()
        assert parquet_path.suffix == ".parquet"

    # Cleanup
    extracted_db.unlink()


def test_full_workflow_from_tar_gz_to_parquets(chembl_tar_gz, temp_base_dir):
    """Test the complete workflow: extract tar.gz -> convert to parquets."""
    # Step 1: Extract the database
    extracted_db = ensure_untarred_sqlite(chembl_tar_gz)
    assert extracted_db.exists()

    # Step 2: Convert to parquets
    parquet_dir = temp_base_dir / "clean"
    parquet_paths = ensure_sqlite_db_to_parquets(extracted_db, cache_dir=parquet_dir)

    # Step 3: Verify all parquets exist
    assert len(parquet_paths) == 4  # We created 4 tables

    # Step 4: Verify parquet content with polars
    import polars as pl

    molecules_lf = pl.scan_parquet(parquet_paths["molecule_dictionary"])
    molecules: pl.DataFrame = molecules_lf.collect()  # ty: ignore[invalid-assignment]

    assert molecules.height == 2
    assert "molregno" in molecules.columns
    assert "chembl_id" in molecules.columns
    assert "pref_name" in molecules.columns

    activities_lf = pl.scan_parquet(parquet_paths["activities"])
    activities: pl.DataFrame = activities_lf.collect()  # ty: ignore[invalid-assignment]

    assert activities.height == 2
    assert "activity_id" in activities.columns
    assert "standard_value" in activities.columns

    # Cleanup
    extracted_db.unlink()


def test_parquet_files_use_separate_temp_dir(chembl_tar_gz, temp_base_dir):
    """Test that parquet conversion doesn't pollute the original temp_base_dir."""
    extracted_db = ensure_untarred_sqlite(chembl_tar_gz)

    parquet_dir_1 = temp_base_dir / "clean_1"
    parquet_paths_1 = ensure_sqlite_db_to_parquets(
        extracted_db, cache_dir=parquet_dir_1
    )

    parquet_dir_2 = temp_base_dir / "clean_2"
    parquet_paths_2 = ensure_sqlite_db_to_parquets(
        extracted_db, cache_dir=parquet_dir_2
    )
    assert parquet_dir_1 != parquet_dir_2
    assert parquet_dir_1.exists()
    assert parquet_dir_2.exists()

    # Verify both have the same tables but in different locations
    assert set(parquet_paths_1.keys()) == set(parquet_paths_2.keys())

    # Verify no cross-contamination
    for path in parquet_paths_1.values():
        assert "clean_1" in str(path)

    for path in parquet_paths_2.values():
        assert "clean_2" in str(path)

    # Cleanup
    extracted_db.unlink()


def test_multiple_databases_in_separate_temp_dirs(temp_base_dir):
    """Test handling multiple databases in separate temporary directories."""
    temp_dirs = []

    for i in range(3):
        # Create a temporary directory for each database
        temp_dir = Path(tempfile.mkdtemp())
        temp_dirs.append(temp_dir)

        # Create a simple database
        db_path = temp_dir / f"test_{i}.db"
        conn = sqlite3.connect(db_path)
        cursor = conn.cursor()
        cursor.execute(f"CREATE TABLE table_{i} (id INTEGER PRIMARY KEY, value TEXT)")
        cursor.execute(f"INSERT INTO table_{i} VALUES (1, 'value_{i}')")
        conn.commit()
        conn.close()

        # Convert to parquets
        parquet_dir = temp_dir / "parquets"
        parquet_paths = ensure_sqlite_db_to_parquets(db_path, cache_dir=parquet_dir)

        # Verify conversion
        assert len(parquet_paths) == 1
        assert parquet_paths[f"table_{i}"].exists()

    # Cleanup all temp directories
    for temp_dir in temp_dirs:
        if temp_dir.exists():
            shutil.rmtree(temp_dir)



"""
Molecule standardisation pipeline for Polars LazyFrames.

Design:
- Dedup on raw SMILES (string level) before any RDKit work
- Multiprocessing for mol parsing, standardisation, and representation
- Lazy Polars throughout except for the unavoidable collect at dedup boundary
- Re-join results back to original frame (preserving all rows)
"""

from __future__ import annotations

import logging as lg
from collections.abc import Callable
from concurrent.futures import ProcessPoolExecutor
from functools import partial
from multiprocessing import get_context as _get_mp_context
from typing import Any, Sequence, cast

import polars as pl
from rdkit.Chem import Mol

from fairfetched.utils._track import track

from ._types import MolToStrFn, RepFns, StdFn, StrToMolFn, T_in, T_out

# Spawn context: safe with JAX (no fork)
_SPAWN_CONTEXT_MP = _get_mp_context("spawn")


def _apply_str_to_mol(molstring: str, read_fn: StrToMolFn) -> tuple[str, Mol | None]:
    try:
        return molstring, read_fn(molstring)
    except Exception:
        return molstring, None


def _apply_standardisation_fns(
    str_mol_tuple: tuple[str, Mol | None], std_fns: list[StdFn]
) -> tuple[str, Mol | None]:
    molstring, mol = str_mol_tuple
    if mol is None:
        return molstring, None
    for fn in std_fns:
        try:
            mol = fn(mol)
        except Exception as e:
            lg.error(f"failed standardisation of {molstring} with exception: {e}")
            mol = None
        if mol is None:
            lg.error(f"standardisation of {molstring} returned None")
            return molstring, None
    return molstring, mol


def _apply_mol_to_reprs(
    str_mol_tuple: tuple[str, Mol | None], rep_fns: dict[str, MolToStrFn]
) -> tuple[str, dict[str, Any | None]]:
    og_string, mol = str_mol_tuple
    if mol is None:
        return og_string, {col: None for col in rep_fns}
    results = {}
    for col, fn in rep_fns.items():
        try:
            results[col] = fn(mol)
        except Exception as e:
            lg.error(f"failed to generate {col} for {og_string} with exception {e}")
            results[col] = None
    return og_string, results


def _imap_mp(
    fn: Callable[[T_in], T_out],
    items: Sequence[T_in],
    n_workers: int,
    chunksize: int,
    desc: str | None = None,
) -> list[T_out]:
    """
    Spawn-safe parallel map via ProcessPoolExecutor.
    Submits chunks as single futures to reduce IPC overhead.
    Results are returned in original order.
    tqdm tracks chunk completion (not individual mols, to avoid IPC per-item).
    """
    if not items:
        return []

    with ProcessPoolExecutor(
        max_workers=n_workers, mp_context=_SPAWN_CONTEXT_MP
    ) as executor:
        result = track(
            executor.map(fn, items, chunksize=chunksize), total=len(items), desc=desc
        )
        return list(result)


# ── steps ─────────────────────────────────────────────────────────────


def _imap_mp_standardisation_fns(
    parsed: list[tuple[str, Mol | None]],
    std_fns: list[StdFn],
    n_workers: int,
    chunksize: int,
) -> list[tuple[str, Mol | None]]:
    """Apply standardisation functions sequentially per mol, in parallel across mols."""
    if not std_fns:
        return parsed
    fn = partial(_apply_standardisation_fns, std_fns=std_fns)
    return _imap_mp(fn, parsed, n_workers, chunksize, desc="standardising")


def build_result_frame(
    results: list[tuple[str, dict[str, Any | None]]],
    molstring_col: str,
) -> pl.LazyFrame:
    if not results:
        return pl.LazyFrame({molstring_col: []})
    og_molstring_list, repr_dicts = zip(*results)
    data: dict[str, list] = {molstring_col: list(og_molstring_list)}
    for col in repr_dicts[0]:
        data[col] = [d[col] for d in repr_dicts]
    return pl.LazyFrame(data)


def add_results(
    lf: pl.LazyFrame,
    results: list[tuple[str, dict[str, Any | None]]],
    molstring_col: str,
) -> pl.LazyFrame:
    if not results:
        return lf
    og_molstring_list, repr_dicts = zip(*results)
    data: dict[str, list] = {molstring_col: list(og_molstring_list)}
    for col in repr_dicts[0]:
        data[col] = [d[col] for d in repr_dicts]
    return (
        lf.with_row_index()
        .join(pl.LazyFrame(data), on=molstring_col, how="left")
        .sort("index")
        .drop("index")
    )


def standardise_mol(
    lf: pl.LazyFrame,
    str_to_mol_fn: StrToMolFn,
    std_fns: list[StdFn],
    mol_to_repr_fns: RepFns,
    molstring_col: str = "smiles",
    output_col: str = "representation",  # ignored when dict passed
    suffix: str = "_original",
    workers: int | None = None,
    chunksize: int = 256,
) -> pl.LazyFrame:
    """
    Standardise a lazyframe

    Args:
        lf:         pl.LazyFrame
        read_fn:    str -> Mol | None
        std_fns:    list of Mol -> Mol | None, applied in order
        rep_fn:     Mol -> Any | None (e.g. canonical SMILES, fingerprint)
        molstring_col: input column name
        output_col: output column name
        n_workers:  number of processes (default: cpu_count)
        chunksize:  imap chunksize (tune for mol size vs overhead)

    Returns:
        pl.LazyFrame, collects only unique smiles

    Use:
        either as a function with lf as the first argument,
            ```
            lf = pl.scan_parquet("my_parquet.parquet")
            lf = standardise_mol(lf, *args, **kwargs)
            ```
        or through polar's `pipe`:
            ```
            lf = lf.pipe(standardise_mol, *args, **kwargs)
            ```
    Note:
        str_to_mol_fn, std_fns, and mol_to_repr_fns must be picklable.
        Use module-level Python wrapper functions, not bare RDKit/Boost.Python
        builtins (e.g. use `smiles_to_mol` not `MolFromSmiles` directly, or make a
        ```
        def mol_from_smiles(smi):
            return MolFromSmiles(smi)
        ```).
    """
    if not isinstance(mol_to_repr_fns, dict):
        mol_to_repr_fns: dict[str, MolToStrFn] = {output_col: mol_to_repr_fns}

    workers = workers or _SPAWN_CONTEXT_MP.cpu_count()

    unique_molstrings: list[str] = (
        lf.select(molstring_col)
        .unique()
        .collect()
        .get_column(molstring_col)  # ty: ignore[unresolved-attribute]
        .to_list()
    )
    read_fn = cast(
        Callable[[str], tuple[str, Mol | None]],
        partial(_apply_str_to_mol, read_fn=str_to_mol_fn),
    )
    to_representations_fn = cast(
        Callable[[tuple[str, Mol | None]], tuple[str, dict[str, Any | None]]],
        partial(_apply_mol_to_reprs, rep_fns=mol_to_repr_fns),
    )

    parsed = _imap_mp(read_fn, unique_molstrings, workers, chunksize, desc="parsing")
    standardised = _imap_mp_standardisation_fns(parsed, std_fns, workers, chunksize)
    results = _imap_mp(
        to_representations_fn,
        standardised,
        workers,
        chunksize,
        desc="converting to string representations",
    )
    return lf.pipe(add_results, results, molstring_col)


def standardisation_function(
    str_to_mol_fn: StrToMolFn,
    std_fns: list[StdFn],
    mol_to_repr_fns: RepFns,
    molstring_col: str = "smiles",
    output_col: str = "representation",
    suffix: str = "_original",
    workers: int | None = None,
    chunksize: int = 256,
) -> Callable[[pl.LazyFrame], pl.LazyFrame]:
    """
    Build a standardise() function from composable pieces.

    Args:
        read_fn:    str -> Mol | None
        std_fns:    list of Mol -> Mol | None, applied in order
        rep_fn:     Mol -> Any | None (e.g. canonical SMILES, fingerprint)
        molstring_col: input column name
        output_col: output column name
        n_workers:  number of processes (default: cpu_count)
        chunksize:  imap chunksize (tune for mol size vs overhead)

    Returns:
        standardise(lf: pl.LazyFrame) -> pl.LazyFrame
    """
    workers = workers or _SPAWN_CONTEXT_MP.cpu_count()

    return partial(
        standardise_mol,
        str_to_mol_fn=str_to_mol_fn,
        std_fns=std_fns,
        mol_to_repr_fns=mol_to_repr_fns,
        molstring_col=molstring_col,
        output_col=output_col,
        suffix=suffix,
        workers=workers,
        chunksize=chunksize,
    )
